#ifndef _INCLUDED_MVPlayer_h_
#define _INCLUDED_MVPlayer_h_

#include <types.h>
#define MVPLAYER_VERSION_STRING "2.0.131129.30OC"

#ifndef MVPLAYER_only_define_version	// Define this symbol before including MVPLayer.h in order to just retrieve the version number





struct MVPlayerDecodedImage
{
	const void *y;
	const void *uv;
	int width;
	int height;
	int left;
	int top;
	int right;
	int bottom;
	unsigned long long *submittedTimeStamp;
};
unsigned long long getSubmittedTimeStamp(void);

/**

A Player widget can appear minimized to its controls, typically on the GamePad

*/
enum MVPlayerVisibility
{
	MVPLAYER_VISIBILITY_HIDDEN = 0,
	MVPLAYER_VISIBILITY_HOURGLASS_ONLY = 1,
	MVPLAYER_VISIBILITY_CONTROLS_ONLY = 2,
	MVPLAYER_VISIBILITY_VIDEO_ONLY = 4,
	MVPLAYER_VISIBILITY_VIDEO_AND_HOURGLASS = 5,
	MVPLAYER_VISIBILITY_VIDEO_AND_CONTROLS = 7
};

/**

A View is used to specify where the Player should display itself

*/
class MVView
{
public:
	MVView();

	/// This view is for TV or GamePad
	int targetScreen;

	/// The player may appear differently on this screen, depending on its current display mode
	MVPlayerVisibility visibility;
	/// The player may send sound to this screen (100), or not at all (0)
	int volumePercent;

	/// The builtin shaders support basic 3D rendering
	WKgfx::Matrix transform;
	WKgfx::Matrix projection;

	/// 16/9 by default. Set to 0 to disable letterboxing (black bands) by the engine
	float windowAspectRatio;

	void InitTransform( float x, float y, float width, float height, const float viewXform[4][4] );
	/// Render the current video frame and GUI controls to a screen
	void Draw();

	bool AdvanceFrame();
	bool GetDecodedImage( MVPlayerDecodedImage* image );

	float fadeColor[4];
	float fadeFactor;

	// The player that this view belongs to ; for internal use
	class MVPlayer* owner;
};

/**

The Player may be running in the following states.
Negative values are final (stop) states, whereas states above zero are transient states.

*/
enum MVPlayerState
{
	/// The media playback failed, the user hit the Close (X) button, or called MVPlayer::Abort() 
	MVPLAYER_STATE_ERROR = -2,
	/// The playback reached the end of the media successfully
	MVPLAYER_STATE_COMPLETE = -1,
	/// Playback has not started yet.
	MVPLAYER_STATE_CLOSED = 0,
	MVPLAYER_STATE_PREPARING,
	MVPLAYER_STATE_PAUSED,
	MVPLAYER_STATE_PLAYING,
	MVPLAYER_STATE_BUFFERING,
	MVPLAYER_STATE_SEEKING
};

enum MVPlayerEvent
{
	MVPLAYER_EVENT_BUFFERING_START,
	MVPLAYER_EVENT_BUFFERING_END
};

/**

The Player may return the following errors.
This is inspired by Error codes from HTML5 Media elements

*/
enum MVPlayerError
{
	/// There is no error currently
	MVPLAYER_ERROR_NONE = 0,
	/// The user pressed the Close button, or the API MVPlayer::Abort() was called
	MVPLAYER_ERROR_ABORTED,
	/// 
	MVPLAYER_ERROR_NETWORK,
	MVPLAYER_ERROR_DECODE,
	MVPLAYER_ERROR_SRC_NOT_SUPPORTED
};

/** 

An error code from the subsystem which initially failed (when available)
When adding a value here, please notify #13654

*/
enum MVPlayerErrorOrigins
{
	/// There is no error currently, or we could not determine any additional details about the origin of an error
	MVPLAYER_ERROR_ORIGIN_UNKNOWN = 0,

	/// 1000..+99	= CURLcode: 6 Couldn't Resolve Host ...
	MVPLAYER_ERROR_ORIGIN_CURL_START = 1000, // CURLE_OK,
	MVPLAYER_ERROR_ORIGIN_CURL_END = MVPLAYER_ERROR_ORIGIN_CURL_START + 99,
	/// 1800..+1001	= X509 peer certificate validation status codes: 19 Couldn't Resolve Host ...
	MVPLAYER_ERROR_ORIGIN_CERT_START = 1800, // NSSL_CERT_VALID
	MVPLAYER_ERROR_ORIGIN_CERT_END = MVPLAYER_ERROR_ORIGIN_CERT_START + 1001, // NSSL_CERT_V_ERR_HOSTNAME_MISMATCH

	/// 5000..+599	= HTTP status code: 404 Not Found ; 403 Forbidden ; 500 Internal Server Error ...
	MVPLAYER_ERROR_ORIGIN_HTTP_START = 5000,
	MVPLAYER_ERROR_ORIGIN_HTTP_END = MVPLAYER_ERROR_ORIGIN_HTTP_START + 599,

	/// 6000	 	= format not supported
	MVPLAYER_ERROR_ORIGIN_FORMAT = 6000,
	MVPLAYER_ERROR_ORIGIN_MEMORY = 6012,

	/// 0x01000000..0x01180000	= H264Dec error: 0x01020000 Out of Memory ; 0x01060000 Mem Segmentation ; 0x01100000 No Data
	MVPLAYER_ERROR_ORIGIN_H264_START = 0x01000000,	// H264DEC_ERR_STREAM
	MVPLAYER_ERROR_ORIGIN_H264_END = 0x01180000,	// H264DEC_ERR_NOTSUPPORTED
};


/**

A Player object is used to stream, decode and render a movie.
As a GUI widget, it can display the video as well as controls for the user to Pause/Resume,
Seek, and Toggle between TV and GamePad.

*/
class MVPlayer
{
public:
	/// Call this function to allocate and initialize instances of the player class
	static MVPlayer* Create();
	/// Call delete on the player instance to stop playback and free resources
	static void Destroy( MVPlayer* player );

	/// Register the URL to stream from (http://example.com/live.m3u8)
	virtual int SetSource( const char* url ) = 0;

	/// Begin downloading and parsing meta data for the stream, asynchronously. When ready for playback, pause and send notification
	virtual int Prepare() = 0;

	/// Resume playback
	virtual int Play() = 0;

	/// Prepare() and Play() asynchronously
	virtual int Start() = 0;

	/// Hold playback. The current frame remains displayed. Returns negative value when cannot perform (player is seeking).
	virtual int Pause() = 0;

	/// Stop playback and free resources
	virtual void Abort() = 0;

	/// Jump to a specific time in the movie, for example to resume.
	virtual int SeekTo( float mediaTimeSecs ) = 0;

	/// Make the GUI react to touch panel input from the user. Provide coordinates calibrated in 1280x720. Returns false if the event was outside the control bar (and not dragging)
	virtual bool ProcessTouchEvent( const void* eventData ) = 0;

	/// Retrieve the current playback position
	virtual float GetCurrentTimeSecs() = 0;

	/// Return the media segment currently preloaded in memory, in seconds
	virtual void GetBufferedRange( float& startSecs, float& endSecs ) = 0;

	/// Configure speed of Playback. 0.5=half, 2.0=double
	virtual int SetPlaybackRate( float multiplier ) = 0;

	/// When rendering to several MVView's (GamePad + TV) in the same screen refresh, you should call this to avoid "wobbling image" bugs
	virtual void ReuseSameVideoFrameForNextDraw() = 0;

	/// Retrieve the current speed of Playback. 0.5=half, 2.0=double
	virtual float GetPlaybackRate() = 0;

	/// SeekTo(), Prepare() and Play() asynchronously
	virtual int Start( float mediaTimeSecs ) = 0;

	virtual bool IsHLS() = 0;

	/// Register callbacks which will be notified when this player instance changes state
	class MVPlayerObserver* observer;

	/// All member variables below are const values, exposed for your convenience
	const char* mediaURL;
	float mediaDurationSecs;

	int videoWidth;
	int videoHeight;

	int audioChannels;
	int audioFrequency;

	/// Playback information
	MVPlayerState state;
	MVPlayerState prevState;

	/// Error information
	MVPlayerError error;
	MVPlayerErrorOrigins errorOrigin;

	/// How the player widget is configured to render itself on each screen
	MVView viewGamePad;
	MVView viewTV;

	/// 0: unknown  1:mp4  2:HLS
	int streamType;
	virtual int GetEstimatedBandwidthKbps();

	/// Call this function to free remaining global resources after all player instances have been deleted
	static void Shutdown();
	/// Call this optional function to enable debug log channels
	static void EnableLogs( const char* channelNames );
	static void SetLogOptions( const char* iniSettings );

	/// Return a guess whether the library will be able to play a given stream. Returns 0 for "not supported", 1-50 for "maybe" and 51-100 for "probably"
	static int CanPlayType( const char* mime, const char* codecs );

	/// Enumerate MIME formats the engine can play, e.g. "video/mp4", "vnd.apple.mpegurl" etc...
	static int GetSupportedTypesCount();
	static const char* GetSupportedType( int index );

	/// Map a file extension such as "m3u8" to a MIME type "vnd.apple.mpegurl"
	static const char* GetTypeFromExtension( const char* fileExtension );

	/// Return the size of the media segment currently loaded in memory
	virtual unsigned GetBytesLoaded() = 0;

protected:
	MVPlayer();
	virtual ~MVPlayer();
};

/**

The Player Observer interface receives notifications during a playback session.
Derive from this class and register your handler in MVPlayer::SetObserver()

*/
class MVPlayerObserver
{
public:
	/// Called when the user pressed Play/Pause or the player rebuffers
	virtual void OnStateChange( MVPlayer* from, MVPlayerState newState );

	virtual void OnEvent( MVPlayer* from, int event );

	/// Called repeatedly after MVPlayer::Start() while the movie preloads
	virtual void OnBufferingUpdate( MVPlayer* from, float startSecs, float endSecs );

	/// Called every time a new video frame should be rendered. It is safe to block
	// in this thread, or call MVView::Draw()
	virtual void OnFrameUpdate( MVPlayer* from );
};

class MVHttpHook
{
public:
	/// Called before curl_perform() for a HTTP request. Modify any CURL option (proxy, cookies, authentication...) and return 0 to proceed
	virtual int OnHttpRequest( CURL* easyRequest, const char* uri );
	/// Called before curl_perform() for a HTTP request. Modify any CURLM option (sockets cache...) and return 0 to proceed
	virtual int OnMultiConfig( CURLM* multiRequest, const char* uri );
	/// Called when an error was detected on a HTTP request managed with curl. Call curl_multi_info_read() and/or curl_easy_info()
	virtual int OnHttpResponse( CURLM* multi, CURL* easyRequest, const char* uri, curl_socket_t fd );

	/// Utility function which calls the OS's curl_easy_setop() (fixes link difficulties with libncurl.a)
	static CURLcode Call_curl_easy_setopt( CURL *curl, CURLoption option, const char* s );
	static CURLcode Call_curl_easy_setopt( CURL *curl, CURLoption option, long n );
	static CURLcode Call_curl_global_init_mem();
};

class MVThreadHook
{
public:
	/// Called to overload OSCreateThread()
	virtual BOOL CreateThread( struct OSThread* thread, int (*entryPoint)(int intArg, void *ptrArg), int intArg, void* ptrArg, void* stack, unsigned int stackSize, int priority, unsigned short attr );
	/// Called from a child thread which starts up
	virtual int OnThreadStart( struct OSThread* th );
	/// Called from a child thread which ends up
	virtual void OnThreadEnd( struct OSThread* th );

	virtual struct OSThread* AllocThread();
	virtual void FreeThread( struct OSThread* allocated );
};

class MVLoadHook
{
public:
	/// @return Total size of the file in bytes, or a negative error
	virtual long long GetFileSize(int fd);

	/// @return A non-negative file descriptor, or a negative error.
	virtual int OpenURI(const char *uri);

	/// @return Number of bytes read successfully.
	virtual int Read(int fd, long long offset, void *buffer, int count);

	virtual void Close(int fd);
};

extern MVHttpHook* MV_HttpHook;
extern MVThreadHook* MV_ThreadHook;
extern MVLoadHook* MV_LoadHook;

/**

To control all memory allocations done by the MediaPlayer engine, you can provide your own heap
in the MV_Heap global variable, before calling MVPlayer::Create().
By default, MVPlayer will use MEMGetBaseHeapHandle(MEM_ARENA_2).

The address range must NOT cross a 256MB boundary.

Alternatively, you can override all memory functions by setting the global pointers below.

*/

extern void* (*MV_malloc_ptr)( size_t size );
extern void* (*MV_realloc_ptr)( void* p, size_t size );
extern void  (*MV_free_ptr)( void* p );
extern void* (*MV_malloc_aligned_ptr)( size_t size, int alignment );
extern void (*MV_free_aligned_ptr)( void* p );
extern void* (*MV_malloc_safe_ptr)( size_t size, int alignment, size_t remaining );
extern "C"
{
	// The functions below call MV_malloc(). Handy for curl_init_memory()
	void* MV_calloc(size_t, size_t);
	char* MV_strdup(const char *);
}

/// If you provide this separate heap, each instance of the H264 decoder will allocate 39 MB from it  The address range must NOT cross a 256MB boundary

/// Change the supported specs of the H264 decoder. These affect memory requirements. Default values are provided
extern int MV_H264Profile; // eg 66=Baseline 77=Main 100=High
extern int MV_H264Level; // eg 42 for "Level 4.2"
extern int MV_H264Width; // eg 1920
extern int MV_H264Height; // eg 1088
extern int MV_H264Frames; // eg 7

/// Change the maximum size allocated for read-ahead buffers (both MP4 and HLS) in bytes
extern int MV_CacheSize;	// eg 20*1024*1024


#endif
#endif
